/*  main_xt.c

    Main file for  krenzo  (X11 contour field tool for Karma).

    Copyright (C) 1996-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive viewing of a contour field
    (otherwise known as a "Renzogram", after Renzo Sanchissi).
    This module runs on an X11 server.


    Written by      Richard Gooch   4-SEP-1996: Copied from kpvslice module.

    Updated by      Richard Gooch   6-SEP-1996: Added start, end and skip
  channel sliders. Do not set levels if level value has not changed.

    Updated by      Richard Gooch   11-SEP-1996: Invalidate old level when new
  cube is loaded.

    Updated by      Richard Gooch   12-SEP-1996: Make sure after loading cube
  <change_active_states> is called after globals are updated.

    Updated by      Richard Gooch   14-SEP-1996: Made use of MomentGenerator
  widget.

    Updated by      Richard Gooch   16-SEP-1996: Made use of TracePopup
  widget.

    Updated by      Richard Gooch   19-SEP-1996: Changed to special function
  call interface for setting new array for MomentGenerator and TracePopup.

    Updated by      Richard Gooch   26-SEP-1996: Fixed display of cube maximum.

    Updated by      Richard Gooch   1-OCT-1996: Fixed resource names for new
  ExportMenu widget. Added "0th channel" image mode.

    Updated by      Richard Gooch   28-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   3-NOV-1996: Made use of new
  <viewimg_track _compute> routine.

    Updated by      Richard Gooch   1-DEC-1996: Made use of
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   27-DEC-1996: Made use of <ds_event_*>
  system. Switched to <dmx> package. Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   30-DEC-1996: Took account of another change
  to <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   31-DEC-1996: Took account of change to
  to Dataclip widget.

    Updated by      Richard Gooch   12-JAN-1997: Made use of new
  XkwNtrapDirectoryDatasets resource for FilewinWidgetClass.

    Updated by      Richard Gooch   27-JAN-1997: Took account of change to
  interface for <XkwTracePopupShowTrace> function.

    Updated by      Richard Gooch   28-JAN-1997: Took account of change back to
  previous interface for <XkwTracePopupShowTrace> function.

    Updated by      Richard Gooch   12-FEB-1997: Only use GLISH "image" and
  "cube" events.

    Updated by      Richard Gooch   9-MAR-1997: Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   26-MAR-1997: Track display now also updates
  when dragging zoom box.

    Updated by      Richard Gooch   19-APR-1997: Now use internal file loader
  in Filewinpopup widget.

    Updated by      Richard Gooch   18-JUN-1997: Support TrueColour-only
  display.

    Updated by      Richard Gooch   22-JUN-1997: Took account of change to
  <xtmisc_init_app_initialise> to double varargs list.

    Updated by      Richard Gooch   12-AUG-1997: No longer initialise <contour>
  package because ImageDisplay widget does it.

    Updated by      Richard Gooch   19-AUG-1997: Set resources in create call
  for second fileselector.

    Updated by      Richard Gooch   18-SEP-1997: Account for data scaling in
  contour level.

    Updated by      Richard Gooch   1-OCT-1997: Took account of change to
  <contour_set_levels>.

    Updated by      Richard Gooch   31-OCT-1997: Took account of change to
  <XkwDataclipNewArray>.

    Updated by      Richard Gooch   7-DEC-1997: Added font resource for new
  intensity menu and removed code to set pan centre for magnifier as it is now
  done automatically by the MagnifierPopup widget. Took account of change to
  <contour_set_levels>.

    Updated by      Richard Gooch   1-MAR-1998: Made use of new filewinWidget
  resource.

    Updated by      Richard Gooch   2-SEP-1998: Removed overlay code.

    Updated by      Richard Gooch   5-SEP-1998: Used new comms. setup.

    Updated by      Richard Gooch   8-SEP-1998: Made use of new function
  <xc_get_location2>.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package. Use SimpleMenu class name in resources.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   31-AUG-1999: Took account of interface
  change to function <viewimg_track _compute>. Added 'l' key binding in
  function <track_ref_event>.

    Updated by      Richard Gooch   9-SEP-1999: Removed track display code and
  made use of XkwNupdateTrackLabels resource for ImageDisplay widget.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   22-JUN-2000: Update Z profile on all events
  rather than limiting to pointer move events. Invalidate Z profile when event
  lies outside cube boundaries.

    Updated by      Richard Gooch   26-JUN-2000: Do not update Z profile on
  pointer enter events.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kcmap.h>
  and <karma_contour.h>.

    Updated by      Richard Gooch   9-AUG-2001: Select visible canvas at
  initialisation.

    Last updated by Richard Gooch   3-OCT-2001: Added translations for Filewin.


*/
#include <stdio.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#define NEW_WIN_SCALE
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_contour.h>
#include <karma_xtmisc.h>
#include <karma_kcmap.h>
#include <karma_dsxfr.h>
#include <karma_conn.h>
#include <karma_dir.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_hi.h>
#include <karma_xc.h>
#include <karma_ic.h>
#include <karma_ex.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/Cmapwinpopup.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/MomentGenerator.h>
#include <Xkw/TracePopup.h>
#ifdef OS_SunOS
#  include <floatingpoint.h>
#endif
#define NEED_IMAGE_MODE_ALTERNATIVES
#include "krenzo.h"


#define VERSION "1.6.25"

#define DEFAULT_COLOURMAP_NAME "Greyscale1"
#define NUM_CONTOUR_COLOURS (unsigned int) 64


/*  Private functions  */
STATIC_FUNCTION (flag main_canvas_event,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin, unsigned int value_type,
		  double x_im, double y_im, int x_pix, int y_pix) );
STATIC_FUNCTION (void image_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void show_image,
		 (KWorldCanvas canvas, iarray image,ViewableImage vimage) );
STATIC_FUNCTION (void show_mom0, (KWorldCanvas canvas) );
STATIC_FUNCTION (void show_mom1, (KWorldCanvas canvas) );
STATIC_FUNCTION (Widget create_levels_popup, (Widget main_shell) );
STATIC_FUNCTION (void moment_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void apply_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void cmap_resize_func, (Kcolourmap cmap, void **info) );
STATIC_FUNCTION (void channels_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void destroy_func, (iarray array, void *info) );
STATIC_FUNCTION (void trace_worldcanvas_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void trace_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void show_channel0, (KWorldCanvas canvas) );
STATIC_FUNCTION (flag cube_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag image_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


/*  Private data  */
String fallback_resources[] =
{
    "Krenzo.geometry:                                 +0+0",
    "Krenzo*pseudoColourCanvas*background:            black",
    "Krenzo*directColourCanvas*background:            black",
    "Krenzo*trueColourCanvas*background:              black",
    "Krenzo*pseudoColourStereoCanvas*background:      black",
    "Krenzo*directColourStereoCanvas*background:      black",
    "Krenzo*trueColourStereoCanvas*background:        black",
    "Krenzo*AnimateControl*positionCanvas.background: black",
    "Krenzo*pseudoColourCanvas*foreground:            white",
    "Krenzo*Command*background:                       grey70",
    "Krenzo*Repeater*background:                      grey70",
    "Krenzo*Ktoggle*background:                       grey80",
    "Krenzo*closeButton*background:                   grey90",
    "Krenzo*ChoiceMenu.background:                    turquoise",
    "Krenzo*ExclusiveMenu.background:                 turquoise",
    "Krenzo*Value*background:                         #d0a0a0",
    "Krenzo*ImageDisplay*quit*background:             orange",
    "Krenzo*SimpleSlider.foreground:                  Sea Green",
    "Krenzo*SimpleProgressMeter.background:           turquoise3",
    "Krenzo*SimpleProgressMeter.foreground:           grey60",
    "Krenzo*ImageDisplay*trackLabel0*font:            8x13bold",
    "Krenzo*ImageDisplay*trackLabel1*font:            8x13bold",
    "Krenzo*ImageDisplay*trackLabel2*font:            8x13bold",
    "Krenzo*ImageDisplay*trackLabel3*font:            8x13bold",
    "Krenzo*ExclusiveMenu.font:                       8x13bold",
    "Krenzo*SimpleMenu*font:                          10x20",
    "Krenzo*ImageDisplay.cmapSize:                    128",
    "Krenzo*ImageDisplay.Command.font:                7x13bold",
    "Krenzo*ImageDisplay.applicationBox.Command.font: 7x13bold",
    "Krenzo*ImageDisplay.applicationBox.Ktoggle.font: 7x13bold",
    "Krenzo*ImageDisplay.ChoiceMenu.font:             7x13bold",
    "Krenzo*ImageDisplay.exportMenu.menuButton.font:  7x13bold",
    "Krenzo*ImageDisplay.overlayMenu.menuButton.font: 7x13bold",
    "Krenzo*ImageDisplay*zoomMenu*Unzoom*foreground:  red",
    "Krenzo*Dataclip.AutoValueScale:                  False",
    "Krenzo*SimpleSlider.borderWidth:                 0",
    "Krenzo*IntensityPolicy*autoIntensityScale:       False",
    "Krenzo*ZoomPolicy*integerXZoom:                  False",
    "Krenzo*ZoomPolicy*integerYZoom:                  False",
    "Krenzo*Filewin.dirname.textSink.font:            fixed",
    "Krenzo*Filewin*list.translations:           #override<MouseMoved>: Set()",
    "Krenzo*font:                                     9x15bold",
    "Krenzo*borderColor:                              black",
    "Krenzo*background:                               aquamarine",
    "Krenzo*foreground:                               black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XPointer) "True"},
    {"-num_colours", ".topForm.cmapSize", XrmoptionSepArg, (XPointer) NULL},
    {"-fullscreen", ".topForm.fullscreen", XrmoptionNoArg, (XPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static Widget image_display = NULL;
static ViewableImage magnified_image = NULL;
static Widget moment_winpopup = NULL;
static Widget contour_level_dlg = NULL;
static Widget start_sld = NULL;
static Widget end_sld = NULL;
static float old_level = TOOBIG;
static Widget cube_min_label = NULL;
static Widget cube_max_label = NULL;
static Widget trace_winpopup = NULL;
static ViewableImage channel0_vimage = NULL;
static double cube_min = TOOBIG;
static double cube_max = -TOOBIG;


/*  Public functions follow  */

int main (int argc, char **argv)
{
    Kdisplay dpy_handle;
    Kcolourmap main_cmap;
    KWorldCanvas pc_canvas, true_canvas;
    XtAppContext app_context;
    Widget filewin, filepopup, w, box, files_btn, menu_btn, cmap_btn;
    Widget trace_btn;
    Widget cmapwinpopup, levels_winpopup;
    extern KWorldCanvas main_canvas;
    extern Kcolourmap contour_cmap;
    extern char title_name[STRING_LENGTH];
    static char function_name[] = "main";

#ifdef SIGFPE_ABORT
	ieee_handler ("set", "division", SIGFPE_ABORT);
#endif
    /*  Initialise module  */
    im_register_module_name ("krenzo");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Krenzo",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_MIN_CCELLS, 200,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    dsxfr_register_connection_limits (1, -1);
    xtmisc_set_icon (main_shell, ic_write_kimage_icon);
    image_display = XtVaCreateManagedWidget ("topForm",
					     imageDisplayWidgetClass,
					     main_shell,
					     XtNborderWidth, 0,
					     XkwNenableAnimation, FALSE,
					     XkwNnumTrackLabels, 3,
					     XkwNupdateTrackLabels, TRUE,
					     NULL);
    if ( ( w = XtNameToWidget (image_display, "filesButton") ) != NULL )
    {
	/*  Force main "Files" button to be re-labelled as "Cubes"  */
	XtVaSetValues (w,
		       XtNlabel, "Cubes",
		       NULL);
    }
    filepopup = XtNameToWidget (image_display, "filewinPopup");
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filepopup,
		   XkwNfilenameTester, XkwFilewinStandardFileTester_3D,
		   XkwNpreferFloat, TRUE,
		   XkwNdomainName, "CUBE",
		   NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (cube_event_func, NULL);
    /*  Get the application box  */
    box = XtNameToWidget (image_display, "applicationBox");
    /*  Create the Image file selector and popup button  */
    filepopup = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass,
				      main_shell,
				      XtNtitle, "Krenzo Image Selector",
				      XtNx, 0,
				      XtNy, 0,
				      XkwNautoPopdown, True,
				      XkwNfilenameTester,
				      XkwFilewinStandardFileTester_nD,
				      XkwNpreferFloat, TRUE,
				      XkwNdomainName, "IMAGE",
				      XkwNloadFiles, TRUE,
				      NULL);
    files_btn = XtVaCreateManagedWidget ("filesButton", commandWidgetClass,box,
					 XtNlabel, "Images",
					 NULL);
    XtAddCallback (files_btn, XtNcallback, xtmisc_popup_cbk, filepopup);
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (image_event_func, NULL);
    /*  Create the image mode menu  */
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, box,
				 XtNmenuName, "imageModeMenu",
				 XkwNchoiceName, "Image Mode",
				 XkwNnumItems, NUM_IMAGE_MODE_ALTERNATIVES,
				 XkwNitemStrings, image_mode_alternatives,
				 NULL);
    menu_btn = w;
    cmap_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
					XtNlabel, "Contour Colourmap",
					NULL);
    /*  Create the trace window popup button  */
    w = XtVaCreateManagedWidget ("traceButton", commandWidgetClass, box,
				 XtNlabel, "Profile",
				 NULL);
    trace_btn = w;
    cube_min_label = XtVaCreateManagedWidget ("minLabel",labelWidgetClass, box,
					      XtNborderWidth, 0,
					      XtNlabel,
					      "Cube min:                  ",
					      NULL);
    cube_max_label = XtVaCreateManagedWidget ("maxLabel",labelWidgetClass, box,
					      XtNborderWidth, 0,
					      XtNlabel,
					      "Cube max:                  ",
					      NULL);
    XtRealizeWidget (main_shell);
    XtVaGetValues (image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   NULL);
    main_canvas = (pc_canvas == NULL) ? true_canvas : pc_canvas;
    if (main_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XtVaSetValues (image_display,
		   XkwNvisibleCanvas, main_canvas,
		   NULL);
    XtAddCallback (menu_btn, XkwNselectCallback, image_mode_cbk, main_canvas);
    viewimg_register_position_event_func (main_canvas,
					  ( flag (*) () ) main_canvas_event,
					  (void *) image_display);
    /*  Create colourmap and control widget for contours  */
    main_cmap = canvas_get_cmap (main_canvas);
    kcmap_get_attributes (main_cmap,
			  KCMAP_ATT_DPY_HANDLE, &dpy_handle,
			  KCMAP_ATT_END);
    /*  Initialise colourmap  */
    if (pc_canvas == NULL)
    {
	unsigned long red, green, blue;

	kwin_get_attributes (canvas_get_pixcanvas (true_canvas),
			     KWIN_ATT_PIX_RED_MASK, &red,
			     KWIN_ATT_PIX_GREEN_MASK, &green,
			     KWIN_ATT_PIX_BLUE_MASK, &blue,
			     KWIN_ATT_END);
	contour_cmap = kcmap_create_scmap (DEFAULT_COLOURMAP_NAME,
					   NUM_CONTOUR_COLOURS, TRUE,
					   red, green, blue);
    }
    else
    {
	contour_cmap = kcmap_va_create (DEFAULT_COLOURMAP_NAME,
					NUM_CONTOUR_COLOURS, TRUE,
					dpy_handle,
					xc_alloc_colours, xc_free_colours,
					xc_store_colours, xc_get_location2,
					KCMAP_ATT_END);
    }
    if (contour_cmap == NULL )
    {
	fprintf (stderr, "Error creating main colourmap\n");
	a_prog_bug (function_name);
    }
    kcmap_register_resize_func (contour_cmap, cmap_resize_func, NULL);
    cmapwinpopup = XtVaCreatePopupShell ("cmapwinpopup",
					 cmapwinpopupWidgetClass, main_shell,
					 XtNtitle, "Contour Colour Control",
					 XkwNkarmaColourmap, contour_cmap,
					 XkwNsimpleColourbar, True,
					 NULL);
    XtAddCallback (cmap_btn, XtNcallback, xtmisc_popup_cbk, cmapwinpopup);
    moment_winpopup = XtVaCreatePopupShell ("orientwinpopup",
					    momentGeneratorWidgetClass,
					    main_shell,
					    XtNtitle,
					    "Moment Generator Control Window",
					    NULL);
    XtAddCallback (moment_winpopup, XkwNmomentCallback, moment_cbk, NULL);
    XtRealizeWidget (moment_winpopup);
    levels_winpopup = create_levels_popup (main_shell);
    XtPopup (levels_winpopup, XtGrabNone);
    trace_winpopup = XtVaCreatePopupShell ("tracewinpopup",
					   tracePopupWidgetClass, main_shell,
					   XtNtitle, "Profile Window",
					   XkwNkarmaColourmap, contour_cmap,
					   NULL);
    XtAddCallback (trace_btn, XtNcallback, xtmisc_popup_cbk, trace_winpopup);
    XtAddCallback (trace_winpopup, XkwNrealiseCallback,
		   trace_worldcanvas_realise_cbk, NULL);    
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */


/*  Private routines follow  */

static flag main_canvas_event (ViewableImage vimage, double x, double y,
			       void *value, unsigned int event_code,
			       void *e_info, void **f_info,
			       double x_lin, double y_lin,
			       unsigned int value_type,
			       double x_im, double y_im, int x_pix, int y_pix)
/*  [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see  canvas_register_transform_func  ) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KwcsAstro main_ap, cube_ap;
    double cube_x, cube_y, min, max;
    dim_desc *dim;
    uaddr coords[2];
    unsigned int dim_indices[2];
    extern iarray cube_arr;
    /*static char function_name[] = "main_canvas_event";*/

    if (event_code == K_CANVAS_EVENT_POINTER_ENTER) return (FALSE);
    if (cube_arr == NULL) return (FALSE);
    cube_ap = iarray_get_associated_wcs_astro (cube_arr);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_WCS_ASTRO, &main_ap,
			    VIEWIMG_VATT_END);
    convert_lin_world_coords (&cube_x, &cube_y, cube_ap,
			      &x_lin, &y_lin, main_ap, 1);
    dim_indices[0] = 2;
    dim_indices[1] = 1;
    dim = iarray_get_dim_desc (cube_arr, 2);
    if (dim->first_coord < dim->last_coord)
    {
	min = dim->first_coord;
	max = dim->last_coord;
    }
    else
    {
	min = dim->last_coord;
	max = dim->first_coord;
    }
    if ( (cube_x < min) || (cube_x > max) )
    {
	XkwTracePopupInvalidate (trace_winpopup);
	return (FALSE);
    }
    coords[0] = ds_get_coord_num (dim, cube_x, SEARCH_BIAS_CLOSEST);
    dim = iarray_get_dim_desc (cube_arr, 1);
    if (dim->first_coord < dim->last_coord)
    {
	min = dim->first_coord;
	max = dim->last_coord;
    }
    else
    {
	min = dim->last_coord;
	max = dim->first_coord;
    }
    if ( (cube_y < min) || (cube_y > max) )
    {
	XkwTracePopupInvalidate (trace_winpopup);
	return (FALSE);
    }
    coords[1] = ds_get_coord_num (dim, cube_y, SEARCH_BIAS_CLOSEST);
    XkwTracePopupShowTrace (trace_winpopup, dim_indices, coords);
    return (FALSE);
}   /*  End Function main_canvas_event  */

static void image_mode_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  [SUMMARY] Image mode change callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = (KWorldCanvas) client_data;
    unsigned int code = *(int *) call_data;
    extern ViewableImage image_vimage;
    extern iarray image_arr;
    extern KwcsAstro main_ap, loaded_image_ap, cube_ap;
    extern unsigned int image_mode;
    /*static char function_name[] = "image_mode_cbk";*/

    image_mode = code;
    switch (code)
    {
      case IMAGE_MODE_LOADED:
	show_image (canvas, image_arr, image_vimage);
	main_ap = loaded_image_ap;
	/*XtPopdown (moment_winpopup);*/
	break;
      case IMAGE_MODE_MOM0:
	XtPopup (moment_winpopup, XtGrabNone);
	show_mom0 (canvas);
	main_ap = cube_ap;
	break;
      case IMAGE_MODE_MOM1:
	XtPopup (moment_winpopup, XtGrabNone);
	show_mom1 (canvas);
	main_ap = cube_ap;
	break;
      case IMAGE_MODE_CHANNEL0:
	show_channel0 (canvas);
	main_ap = cube_ap;
	break;
    }
}   /*  End Function image_mode_cbk  */

static void show_image (KWorldCanvas canvas, iarray image,ViewableImage vimage)
/*  [SUMMARY] Show the loaded image.
    <canvas> The world canvas.
    <image> The image array.
    <vimage> The image.
    [RETURNS] Nothing.
*/
{
    double i_min, i_max;
    /*static char function_name[] = "show_image";*/

    if (vimage == NULL) return;
    if (viewimg_get_active (canvas) == vimage) return;
    /*  Must change the canvas intensity scale  */
    if ( !iarray_min_max (image, CONV1_REAL, &i_min, &i_max) )
    {
	fprintf (stderr, "Error computing min-max\n");
    }
    canvas_set_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, i_min,
			   CANVAS_ATT_VALUE_MAX, i_max,
			   CANVAS_ATT_END);
    viewimg_make_active (vimage);
}   /*  End Function show_image  */

static void show_mom0 (KWorldCanvas canvas)
/*  [SUMMARY] Show the 0th moment map.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    double i_min, i_max;
    extern iarray mom0_arr;
    extern ViewableImage mom0_vimage;
    /*static char function_name[] = "show_mom0";*/

    if (mom0_vimage == NULL) return;
    if (viewimg_get_active (canvas) == mom0_vimage) return;
    /*  Must change the canvas intensity scale  */
    if ( !iarray_min_max (mom0_arr, CONV1_REAL, &i_min, &i_max) )
    {
	fprintf (stderr, "Error computing min-max\n");
    }
    if (i_min >= i_max)
    {
	fprintf (stderr, "Computed 0th moment map is blank: ignoring\n");
	return;
    }
    canvas_set_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, i_min,
			   CANVAS_ATT_VALUE_MAX, i_max,
			   CANVAS_ATT_END);
    viewimg_make_active (mom0_vimage);
}   /*  End Function show_mom0  */

static void show_mom1 (KWorldCanvas canvas)
/*  [SUMMARY] Show the 1st moment map.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    double i_min, i_max;
    extern iarray mom1_arr;
    extern ViewableImage mom0_vimage, mom1_vimage;
    /*static char function_name[] = "show_mom1";*/

    if (mom1_vimage == NULL) return;
    if ( (viewimg_get_active (canvas) == NULL) ||
	 (viewimg_get_active (canvas) == mom0_vimage) )
    {
	/*  Must change the canvas intensity scale  */
	if ( !iarray_min_max (mom1_arr, CONV1_REAL, &i_min, &i_max) )
	{
	    fprintf (stderr, "Error computing min-max\n");
	}
	if (i_min >= i_max)
	{
	    fprintf (stderr, "Computed 1st moment map is blank: ignoring\n");
	    return;
	}
	canvas_set_attributes (canvas,
			       CANVAS_ATT_VALUE_MIN, i_min,
			       CANVAS_ATT_VALUE_MAX, i_max,
			       CANVAS_ATT_END);
    }
    viewimg_make_active (mom1_vimage);
}   /*  End Function show_mom1  */

static Widget create_levels_popup (Widget main_shell)
/*  [SUMMARY] Create the level control popup.
    <main_shell> The main shell widget.
    [RETURNS] The popup.
*/
{
    Widget form, btn, popup, skip_sld;
    extern int start_channel, end_channel, channel_skip;

    popup = XtVaCreatePopupShell ("orientwinpopup",
				  topLevelShellWidgetClass, main_shell,
				  XtNtitle, "Contour Level Control Window",
				  NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, popup,
				    XtNborderWidth,0,
				    NULL);
#ifdef DISABLED
    btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				   XtNlabel, "Close",
				   XtNtop, XtChainTop,
				   XtNbottom, XtChainTop,
				   XtNleft, XtChainLeft,
				   XtNright, XtChainRight,
				   XtNwidth, 100,
				   XtNheight, 20,
				   NULL);
    XtAddCallback (btn, XtNcallback, xtmisc_popdown_cbk, popup);
#endif
    btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				   XtNlabel, "Apply Parameters",
				   /*XtNfromHoriz, btn,*/
				   NULL);
    XtAddCallback (btn, XtNcallback, apply_cbk, NULL);
    contour_level_dlg = XtVaCreateManagedWidget ("contourLevel",
						 dialogWidgetClass, form,
						 XtNlabel, "Contour Level",
						 XtNfromVert, btn,
						 XtNvalue, "5e-2",
						 NULL);
    start_sld = XtVaCreateManagedWidget ("startSlider",
					 simpleSliderWidgetClass, form,
					 XtNlabel, "Start Channel",
					 XtNfromVert, contour_level_dlg,
					 XkwNmodifier, 1,
					 XkwNshowRange, TRUE,
					 XkwNvaluePtr, &start_channel,
					 NULL);
    XtAddCallback (start_sld, XkwNvalueChangeCallback, channels_cbk, NULL);
    end_sld = XtVaCreateManagedWidget ("endSlider",
				       simpleSliderWidgetClass, form,
				       XtNlabel, "End Channel",
				       XtNfromVert, start_sld,
				       XkwNmodifier, 1,
				       XkwNshowRange, TRUE,
				       XkwNvaluePtr, &end_channel,
				       NULL);
    XtAddCallback (end_sld, XkwNvalueChangeCallback, channels_cbk, NULL);
    skip_sld = XtVaCreateManagedWidget ("skipSlider",
					simpleSliderWidgetClass, form,
					XtNlabel, "Channel Skip",
					XtNfromVert, end_sld,
					XkwNmodifier, 1,
					XkwNminimum, 1,
					XkwNmaximum, 10,
					XtNvalue, channel_skip,
					XkwNvaluePtr, &channel_skip,
					NULL);
    XtAddCallback (skip_sld, XkwNvalueChangeCallback, channels_cbk, NULL);
    XtRealizeWidget (popup);
    return (popup);
}   /*  End Function create_levels_popup  */

static void moment_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Moment computation callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    extern iarray cube_arr;
    extern KWorldCanvas main_canvas;
    extern ViewableImage mom0_vimage, mom1_vimage;
    extern unsigned int image_mode;
    /*static char function_name[] = "moment_cbk";*/

    if (cube_arr == NULL) return;
    viewimg_register_data_change (mom0_vimage);
    viewimg_register_data_change (mom1_vimage);
    if (image_mode == IMAGE_MODE_MOM0) show_mom0 (main_canvas);
    if (image_mode == IMAGE_MODE_MOM1) show_mom1 (main_canvas);
}   /*  End Function moment_cbk  */

static void apply_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Contour levels change callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    unsigned int count, num_pixels;
    unsigned long pixel_value;
    float tmp, level;
    double d_level, scale, offset;
    char *clip_str;
    CONST char *p;
    unsigned long *pixel_values;
    extern KWorldCanvas main_canvas;
    extern Kcolourmap contour_cmap;
    extern iarray cube_arr;
    extern flag change_active_cimages;
    extern unsigned int num_cimages;
    extern KContourImage *cimages;
    /*static char function_name[] = "apply_cbk";*/

    if (cube_arr == NULL) return;
    iarray_get_data_scaling (cube_arr, &scale, &offset);
    XtVaGetValues (contour_level_dlg,
		   XtNvalue, &clip_str,
		   NULL);
    level = ex_float (clip_str, &p);
    if (level != old_level)
    {
	num_pixels = kcmap_get_pixels (contour_cmap, &pixel_values);
	d_level = (level - offset) / scale;
	for (count = 0; count < num_cimages; ++count)
	{
	    tmp = (float) count / (float) (num_cimages - 1) *
		(float) (num_pixels - 1);
	    pixel_value = pixel_values[(int) tmp];
	    contour_set_levels (cimages[count], 1, &d_level, &pixel_value,
				NULL, NULL);
	}
	old_level = level;
    }
    if (change_active_cimages) change_active_states ();
    kwin_resize (canvas_get_pixcanvas (main_canvas), FALSE, 0, 0, 0, 0);
}   /*  End Function apply_cbk  */

static void cmap_resize_func (Kcolourmap cmap, void **info)
/*  [SUMMARY] This routine registers a change in the size of a colourmap.
    <cmap> The colourmap.
    <info> A pointer to the arbitrary colourmap information pointer.
    [RETURNS] Nothing.
*/
{
    unsigned int count, num_pixels;
    unsigned long pixel_value;
    float tmp, level;
    double d_level, scale, offset;
    char *clip_str;
    CONST char *p;
    unsigned long *pixel_values;
    extern KWorldCanvas main_canvas;
    extern iarray cube_arr;
    extern unsigned int num_cimages;
    extern KContourImage *cimages;
    /*static char function_name[] = "cmap_resize_func";*/

    if (cube_arr == NULL) return;
    iarray_get_data_scaling (cube_arr, &scale, &offset);
    XtVaGetValues (contour_level_dlg,
		   XtNvalue, &clip_str,
		   NULL);
    level = ex_float (clip_str, &p);
    num_pixels = kcmap_get_pixels (cmap, &pixel_values);
    d_level = (level - offset) / scale;
    for (count = 0; count < num_cimages; ++count)
    {
	tmp = (float) count / (float) (num_cimages - 1) *
	    (float) (num_pixels - 1);
	pixel_value = pixel_values[(int) tmp];
	contour_set_levels (cimages[count], 1, &d_level, &pixel_value,
			    NULL, NULL);
    }
    kwin_resize (canvas_get_pixcanvas (main_canvas), FALSE, 0, 0, 0, 0);
}   /*  End Function cmap_resize_func  */

static void channels_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Channels change callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    extern flag change_active_cimages;

    change_active_cimages = TRUE;
}   /*  End Function channels_cbk  */

static void destroy_func (iarray array, void *info)
/*  [SUMMARY] Register destruction of an Intelligent Array.
    <array> The Intelligent Array.
    <info> A pointer to the arbitrary information.
    [RETURNS] Nothing.
*/
{
    ViewableImage *vimage = info;

    viewimg_destroy (*vimage);
    *vimage = NULL;
}   /*  End Function destroy_func  */

static void trace_worldcanvas_realise_cbk (Widget w, XtPointer client_data,
					   XtPointer call_data)
/*  [SUMMARY] Canvas realise callback.
    [PURPOSE] This routine is called when a Canvas widget is realised and the
    KWorldCanvas object is created.
    <w> The Canvas widget that has just been realised.
    <client_data> The client data.
    <call_data> The world canvas.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas worldcanvas = (KWorldCanvas) call_data;

    canvas_register_refresh_func (worldcanvas, trace_refresh_func, NULL);
}   /*  End Function trace_worldcanvas_realise_cbk   */

static void trace_refresh_func (KWorldCanvas canvas, int width, int height,
				struct win_scale_type *win_scale,
				Kcolourmap cmap, flag cmap_resize, void **info,
				PostScriptPage pspage,
				unsigned int num_areas,
				KPixCanvasRefreshArea *areas,
				flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    double left_x, right_x, dummy;

    if ( !canvas_get_colour (canvas, "white", &pixel_value, NULL, NULL, NULL) )
    {
	pixel_value = 0;
    }
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_END);
    /*  Convert left&right x values to non-linear co-ordinates  */
    dummy = 0.0;
    canvas_coords_transform (canvas, 1, &left_x, FALSE, &dummy, FALSE);
    dummy = 0.0;
    canvas_coords_transform (canvas, 1, &right_x, FALSE, &dummy, FALSE);
    canvas_draw_line_p (canvas, left_x, old_level, right_x, old_level,
			pixel_value);
}   /*  End Function trace_refresh_func  */

static void show_channel0 (KWorldCanvas canvas)
/*  [SUMMARY] Show the 0th channel map.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    /*static char function_name[] = "show_channel0";*/

    if (channel0_vimage == NULL) return;
    if (viewimg_get_active (canvas) == channel0_vimage) return;
    /*  Must change the canvas intensity scale  */
    canvas_set_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, cube_min,
			   CANVAS_ATT_VALUE_MAX, cube_max,
			   CANVAS_ATT_END);
    viewimg_make_active (channel0_vimage);
}   /*  End Function show_channel0  */

static flag cube_event_func (void *object, multi_array *multi_desc,
			     CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new cube event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    iarray arr;
    char filename[STRING_LENGTH];
    char txt[STRING_LENGTH];
    extern KWorldCanvas main_canvas;
    extern iarray cube_arr, mom0_arr, mom1_arr;
    extern ViewableImage mom0_vimage, mom1_vimage;
    extern unsigned int num_cimages;
    static char function_name[] = "cube_event_func";

    if (strcmp (domain, "IMAGE") == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "cube") != 0) )
	return (FALSE);
    if (strcmp (domain, "CUBE") == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    if ( ( arr = process_cube (multi_desc, &cube_min, &cube_max) ) == NULL )
	return (FALSE);
    if (cube_arr != NULL) iarray_dealloc (cube_arr);
    cube_arr = arr;
    old_level = TOOBIG;
    XkwMomentGeneratorNewArray (moment_winpopup, cube_arr, cube_min, cube_max);
    XkwTracePopupNewArray (trace_winpopup, cube_arr, cube_min, cube_max);
    if (channel0_vimage != NULL) viewimg_destroy (channel0_vimage);
    if ( ( channel0_vimage = viewimg_create_restr (main_canvas,
						   cube_arr->multi_desc,
						   cube_arr->arr_desc,
						   cube_arr->data,
						   2, 1, 0,
						   0, NULL, NULL) ) == NULL )
    {
	m_abort (function_name, "0th channel ViewableImage");
    }
    if (mom0_vimage != NULL)
    {
	fprintf (stderr,"WARNING: 0th moment ViewableImage not deallocated\n");
    }
    if (mom1_vimage != NULL)
    {
	fprintf (stderr,"WARNING: 1st moment ViewableImage not deallocated\n");
    }
    XtVaGetValues (moment_winpopup,
		   XkwNmom0Array, &mom0_arr,
		   XkwNmom1Array, &mom1_arr,
		   NULL);
    if ( ( mom0_vimage = viewimg_create_from_iarray (main_canvas, mom0_arr,
						     FALSE) ) == NULL )
    {
	m_abort (function_name, "0th moment ViewableImage");
    }
    if ( ( mom1_vimage = viewimg_create_from_iarray (main_canvas, mom1_arr,
						     FALSE) ) == NULL )
    {
	m_abort (function_name, "1st moment ViewableImage");
    }
    iarray_register_destroy_func (mom0_arr, ( flag (*) () ) destroy_func,
				  &mom0_vimage);
    iarray_register_destroy_func (mom1_arr, ( flag (*) () ) destroy_func,
				  &mom1_vimage);
    sprintf (txt, "Cube min: %e", cube_min);
    XtVaSetValues (cube_min_label,
		   XtNlabel, txt,
		   NULL);
    sprintf (txt, "Cube max: %e", cube_max);
    XtVaSetValues (cube_max_label,
		   XtNlabel, txt,
		   NULL);
    XtVaSetValues (start_sld,
		   XtNvalue, 0,
		   XkwNmaximum, num_cimages - 1,
		   NULL);
    XtVaSetValues (end_sld,
		   XtNvalue, num_cimages - 1,
		   XkwNmaximum, num_cimages - 1,
		   NULL);
    /*  Must wait until sliders are updated before changing the active states,
	since the sliders update some global data  */
    change_active_states ();
    return (TRUE);
}   /*  End Function cube_event_func  */

static flag image_event_func (void *object, multi_array *multi_desc,
			      CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new image event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KWorldCanvas canvas, mag_canvas;
    iarray arr;
    double min, max;
    Widget izoomwinpopup, animate_winpopup, magnifier;
    char *ptr;
    char filename[STRING_LENGTH], stripped_filename[STRING_LENGTH];
    char title[STRING_LENGTH];
    extern ViewableImage image_vimage;
    extern iarray image_arr;
    extern char title_name[STRING_LENGTH];

    if (strcmp (domain, "CUBE") == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "image") != 0) )
	return (FALSE);
    if (strcmp (domain, "IMAGE") == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    strcpy (stripped_filename, filename);
    if ( ( ptr = strrchr (stripped_filename, '.') ) != NULL )
    {
	if (strcmp (ptr, ".kf") == 0) *ptr = '\0';
    }
    XtVaGetValues (image_display,
		   XkwNvisibleCanvas, &canvas,
		   XkwNmagnifier, &magnifier,
		   NULL);
    XtVaGetValues (magnifier, XkwNvisibleCanvas, &mag_canvas, NULL);
    izoomwinpopup = XtNameToWidget (image_display, "izoomwinpopup");
    animate_winpopup = XtNameToWidget (image_display, "animatepopup");
    if ( ( arr = process_image (multi_desc, &min, &max, canvas, mag_canvas,
				&image_vimage, &magnified_image) )
	 == NULL ) return (FALSE);
    if (image_arr != NULL) iarray_dealloc (image_arr);
    image_arr = arr;
    sprintf (title, "%s  file: %s\n", title_name, filename);
    XtVaSetValues (main_shell,
		   XtNtitle, title,
		   NULL);
    XtVaSetValues (image_display,
		   XkwNimageName, stripped_filename,
		   NULL);
    XkwDataclipNewArray (izoomwinpopup, image_arr, min, max, TRUE);
    return (TRUE);
}   /*  End Function image_event_func  */
